//constant time box-filter implementation
void boxfilter3dref(double* vol1,double* temp1,double* temp2,double* ones,int hw,int m,int n,int o){
    const int RANGE=hw;
    const int sz=m*n*o;
    for(int k=0;k<o;k++){
        for(int j=0;j<n;j++){
            double sumA=0;
            const double* inRow=vol1+j*m+k*m*n;
            double* outRow=temp2+j*m+k*m*n;
            
            int x = -RANGE;
            for (; x<=RANGE; x++) {
                sumA += inRow[x+RANGE]; ;
                if (x >= 0) outRow[x] = sumA;
            }
            for (; x+RANGE<m; x++) {
                sumA -= inRow[x-RANGE-1];
                sumA += inRow[x+RANGE];
                outRow[x] = sumA;// * k;
            }
            for (; x < m; x++) {
                sumA -= inRow[x - RANGE - 1]; //sum1 -= 1.0f;
                outRow[x] = sumA;///sum1;
            }
        }
    }
    for(int k=0;k<o;k++){
        for(int i=0;i<m;i++){
            double sumA=0;
            int x = -RANGE;
            for (; x<=RANGE; x++) {
                sumA += temp2[i+(x+RANGE)*m+k*m*n];
                if (x >= 0)
                    temp1[i+x*m+k*m*n]=sumA; //output=>temp1 !!!!
            }
            for (; x+RANGE<n; x++) {
                sumA -= temp2[i+(x-RANGE-1)*m+k*m*n];
                sumA += temp2[i+(x+RANGE)*m+k*m*n];
                temp1[i+x*m+k*m*n]= sumA;// * k;
            }
            for (; x < n; x++) {
                sumA -= temp2[i+(x-RANGE-1)*m+k*m*n];
                temp1[i+x*m+k*m*n] = sumA;///sum1;
            }
        }
    }
    for(int j=0;j<n;j++){
        for(int i=0;i<m;i++){
            double sumA=0;
            int x = -RANGE;
            for (; x<=RANGE; x++) {
                sumA += temp1[i+j*m+(x+RANGE)*m*n];
                if (x >= 0)
                    vol1[i+j*m+x*m*n]=sumA*ones[i+j*m+x*m*n];
            }
            for (; x+RANGE<o; x++) {
                sumA -= temp1[i+j*m+(x-RANGE-1)*m*n];
                sumA += temp1[i+j*m+(x+RANGE)*m*n];
                vol1[i+j*m+x*m*n]= sumA*ones[i+j*m+x*m*n];// * k;
            }
            for (; x < o; x++) {
                sumA -= temp1[i+j*m+(x-RANGE-1)*m*n];
                vol1[i+j*m+x*m*n] = sumA*ones[i+j*m+x*m*n];///sum1;
            }
        }
    }
}

vector<uint64_t> jointNNLSfilter(double* weights,double* weights_lsq,double* weights_ncc,double* image1,double* image2,double* mask,int m,int n,int o,int num_atlas,int r,double eps,pair<double,double>& evalTimes){
    timeval time1,time2;
    uint64_t sz=m*n*o;
    uint64_t ch2=num_atlas;
    
    //int* mask_ind=new int[sz];
    vector<uint64_t> inverse_ind;
    uint64_t mask_num=0;
    for(uint64_t i=0;i<sz;i++){
        if(mask[i]>0){ //mask contains locations without unanimous votes
            //mask_ind[i]=mask_num;
            inverse_ind.push_back(i); //store indices of elements in mask
            mask_num++;
        }
    }
    
    printf("mask_num: %d or %d / %d\n",(int)mask_num,(int)inverse_ind.size(),(int)sz);
    gettimeofday(&time1, NULL);
    
    double* filt1=new double[sz];
    
    double* temp1=new double[sz];
    double* temp2=new double[sz];
    double* ones0=new double[sz];
    
    //set-up boxfilter (number of voxels in every box)
    double* ones=new double[sz];
    for(uint64_t i=0;i<sz;i++){
        ones[i]=1.0;
        ones0[i]=1.0;
    }
    boxfilter3dref(ones,temp1,temp2,ones0,r,m,n,o);
    delete ones0;
    for(uint64_t i=0;i<sz;i++){
        ones[i]=1.0/(double)ones[i];
    }
    
    printf("starting with box-filtering\n");
    int varm=ch2%2==1?(ch2*ceil(ch2/2.0)):((ch2+1)*(ch2-2));
    for(int l=0;l<(2+(2*ch2+varm));l++){
        cout<<"="<<flush;
    }
    cout<<"\n";
    
    //calculate local means and variances of both images using boxfilters
    double* mean1=new double[mask_num];
    double* mean2=new double[mask_num*ch2];
    double* var1=new double[mask_num];
    double* var2=new double[mask_num*ch2*ch2];
    double* cov=new double[mask_num*ch2];
    
    //image1
    for(uint64_t i=0;i<sz;i++){
        filt1[i]=image1[i];
    }
    boxfilter3dref(filt1,temp1,temp2,ones,r,m,n,o);
    cout<<"x"<<flush;
    
    for(uint64_t i=0;i<mask_num;i++){
        mean1[i]=filt1[inverse_ind[i]];
    }
    
    for(uint64_t i=0;i<sz;i++){
        filt1[i]=image1[i]*image1[i];
    }
    boxfilter3dref(filt1,temp1,temp2,ones,r,m,n,o);
    cout<<"x"<<flush;
    
    for(int i=0;i<mask_num;i++){
        var1[i]=filt1[inverse_ind[i]]-mean1[i]*mean1[i];
    }
    
    //image2
    for(uint64_t c=0;c<ch2;c++){
        for(uint64_t i=0;i<sz;i++){
            filt1[i]=image2[i+c*sz];
        }
        boxfilter3dref(filt1,temp1,temp2,ones,r,m,n,o);
        cout<<"x"<<flush;
        
        for(uint64_t i=0;i<mask_num;i++){
            mean2[i+c*mask_num]=filt1[inverse_ind[i]];
        }
    }
    
    for(uint64_t c1=0;c1<ch2;c1++){
        for(uint64_t c2=c1;c2<ch2;c2++){
            for(uint64_t i=0;i<sz;i++){
                filt1[i]=image2[i+c1*sz]*image2[i+c2*sz];
            }
            boxfilter3dref(filt1,temp1,temp2,ones,r,m,n,o);
            cout<<"x"<<flush;
            
            for(uint64_t i=0;i<mask_num;i++){
                var2[i+(c1+c2*ch2)*mask_num]=filt1[inverse_ind[i]]-mean2[i+c1*mask_num]*mean2[i+c2*mask_num];
            }
            
        }
    }
    //cross-images
    for(uint64_t c2=0;c2<ch2;c2++){
        for(uint64_t i=0;i<sz;i++){
            filt1[i]=image1[i]*image2[i+c2*sz];
        }
        boxfilter3dref(filt1,temp1,temp2,ones,r,m,n,o);
        cout<<"x"<<flush;
        
        for(uint64_t i=0;i<mask_num;i++){
            cov[i+c2*mask_num]=filt1[inverse_ind[i]]-mean1[i]*mean2[i+c2*mask_num];
        }
    }
    
    gettimeofday(&time2, NULL);
    double timeP1=time2.tv_sec+time2.tv_usec/1e6-(time1.tv_sec+time1.tv_usec/1e6);
    
    printf("\nComputation time boxfilter: %4.3f secs\n",timeP1);
    printf("starting with joint NNLS computation\n");
    int frac=(int)(mask_num/40);
    for(int i=0;i<mask_num;i++){
        if((i%frac)==0){
            cout<<"="<<flush;
        }
    }
    cout<<"\n";
    
    //jointNNLS-calculation voxelwise
 
    MatrixXd A(ch2,ch2);
    VectorXd b(ch2);
    VectorXd x(ch2);
    VectorXd x_lsq(ch2);
    
    double* denomC=new double[ch2];
    
    for(uint64_t i=0;i<mask_num*ch2;i++){
        weights[i]=0.0;
    }
    uint64_t zeros=0; uint64_t nonnegs=0;
    double timeP2=0; double timeP3=0;
    for(uint64_t i=0;i<mask_num;i++){
        
        
        if((i%frac)==0){
            cout<<"x"<<flush;
        }
        
        double denomB=sqrt(var1[i]);
        
        //cross-correlations
        for(uint64_t c2=0;c2<ch2;c2++){
            denomC[c2]=sqrt(var2[i+(c2+c2*ch2)*mask_num]);
            b(c2)=cov[i+c2*mask_num]/max(denomB*denomC[c2],1e-6);
        }
        
        //inverse cc-var2
        for(uint64_t c1=0;c1<ch2;c1++){
            for(uint64_t c2=c1;c2<ch2;c2++){
                A(c1,c2)=var2[i+(c1+c2*ch2)*mask_num]/max(denomC[c1]*denomC[c2],1e-6);
                A(c2,c1)=var2[i+(c1+c2*ch2)*mask_num]/max(denomC[c1]*denomC[c2],1e-6);
                if(c1==c2){
                    A(c1,c2)+=eps;
                }
            }
        }
        
        gettimeofday(&time1, NULL);

        
        //max_iter, eps
        NNLS<MatrixXd> nnls(A,30,1e-4);
        
        nnls.solve(b);
        x=nnls.x();
        
        gettimeofday(&time2, NULL);
        timeP2+=time2.tv_sec+time2.tv_usec/1e6-(time1.tv_sec+time1.tv_usec/1e6);

        
        gettimeofday(&time1, NULL);

        x_lsq=A.householderQr().solve(b);
        
        gettimeofday(&time2, NULL);
        timeP3+=time2.tv_sec+time2.tv_usec/1e6-(time1.tv_sec+time1.tv_usec/1e6);
        
        double sum_lsq=0.0; double sum_ncc=0.0; double sum_nlss=0.0;
        for(uint64_t c1=0;c1<ch2;c1++){
            double val_nlss=(x[c1]==x[c1])?x[c1]:0.0;
            sum_nlss+=val_nlss;
            sum_lsq+=x_lsq[c1];
            sum_ncc+=b[c1];
        }
        
        
        for(uint64_t c1=0;c1<ch2;c1++){
//            weights[inverse_ind[i]+c1*sz]=x[c1]; //write in actual image dimensions
            double val_nlss=(x[c1]==x[c1])?x[c1]:0.0;
            if(val_nlss==0){ zeros++;}else{ nonnegs++; }
            weights[i+c1*mask_num]=val_nlss;//sum_nlss>0.0?val_nlss/sum_nlss:0.0; //nonnegative least squares
            double val_lsq=x_lsq[c1]==x_lsq[c1]?x_lsq[c1]:0.0;
            double val_ncc=b[c1]==b[c1]?b[c1]:0.0;
            weights_ncc[i+c1*mask_num]=val_ncc;
            //sum_ncc>0.0?val_ncc/sum_ncc:0.0;//ncc (sum to 1)
            weights_lsq[i+c1*mask_num]=val_lsq;
            //sum_lsq>0.0?val_lsq/sum_lsq:0.0;//lsq (sum to 1)

        }
        
        
    }
    
    char percent[3];
    strcpy(percent,"%");

    evalTimes.first=timeP1;
    evalTimes.second=timeP2;
    
    printf("\nComputation time NNLS: %4.3f secs and LSQ: %4.3f secs\n",timeP2,timeP3);
    printf("Sparsity: %d / %d or %4.3f %s\n",(int)zeros,(int)nonnegs,100.0*zeros/(double)(nonnegs+zeros),percent);

    delete filt1;
    delete mean1; delete mean2;
    delete var1; delete var2; delete cov;
    delete temp1; delete temp2; delete ones;
    return inverse_ind;
}


//    jointCCAfilter(similarity,image1,image2,mask,m,n,o,ch2,r,eps);



